package org.embulk.input.filename;

import java.util.List;
import java.util.Arrays;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.nio.file.attribute.BasicFileAttributeView;
import java.nio.file.attribute.BasicFileAttributes;
import java.nio.file.attribute.FileTime;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.nio.file.Path;
import java.nio.file.SimpleFileVisitor;
import java.nio.file.FileVisitResult;
import java.io.File;
import java.io.IOException;
import java.io.FileInputStream;
import java.io.ByteArrayOutputStream;

import com.google.common.base.Optional;
import org.apache.commons.codec.binary.Base64;

//import org.apache.commons.io.IOUtils;

import org.slf4j.Logger;


import org.embulk.config.Config;
import org.embulk.config.ConfigDefault;
import org.embulk.config.ConfigDiff;
import org.embulk.config.ConfigSource;
import org.embulk.config.Task;
import org.embulk.config.TaskReport;
import org.embulk.config.TaskSource;
import org.embulk.config.ConfigInject;
import org.embulk.spi.PageBuilder;
import org.embulk.spi.Exec;
import org.embulk.spi.InputPlugin;
import org.embulk.spi.PageOutput;
import org.embulk.spi.Schema;
import org.embulk.spi.SchemaConfig;
import org.embulk.spi.BufferAllocator;
import org.embulk.spi.ColumnConfig;
import org.embulk.spi.SchemaConfig;

import static org.embulk.spi.type.Types.STRING;


public class FilenameInputPlugin
        implements InputPlugin
{
    public interface PluginTask
            extends Task
    {
        @Config("multi_dir")
        @ConfigDefault("[]")
        ArrayList<String> getMultiDir();
         
        @Config("multi_tag")
        @ConfigDefault("[]")
        ArrayList<String> getMultiTag();
        
        @Config("lastPaths")
        @ConfigDefault("[]")
        ArrayList<String> getLastPaths();
        
        @Config("order")
        @ConfigDefault("ALPHABETICAL")
        String getOrder();
        
        @Config("chunk_size")
        @ConfigDefault("10485760")
        int getChunkSize();



        @Config("follow_symlinks")
        @ConfigDefault("false")
        boolean getFollowSymlinks();
        

        ArrayList<ArrayList<String>> getFiles();
        void setFiles(ArrayList<ArrayList<String>> allFiles);

        @ConfigInject
        BufferAllocator getBufferAllocator();
       
       
    }
    
    private final Logger log = Exec.getLogger(getClass());

    private final static Path CURRENT_DIR = Paths.get(".").normalize();
    private static ArrayList<String> tagList;
    private static ArrayList<String> lastPaths;
    
    private static int chunkSize;

    @Override
    public ConfigDiff transaction(ConfigSource config,
            InputPlugin.Control control)
    {
        PluginTask task = config.loadConfig(PluginTask.class);
        
        chunkSize = task.getChunkSize();
        ArrayList<String> dirList = task.getMultiDir();
        ArrayList<String> lastPaths = task.getLastPaths();
        
        ArrayList<ArrayList<String>> allFiles = new ArrayList<ArrayList<String>>();
        tagList = task.getMultiTag();
        if ( dirList.size() != 0 ){
            log.info ("The list of the directories: " + dirList );
            while (tagList.size() < dirList.size()){
                // If the Number of tags is less than the directories, we say that the default tag is ""
                tagList.add("");
            }
            while (lastPaths.size() < dirList.size()){
                lastPaths.add("");
            }
        } else {
            throw new RuntimeException("The multi_dir should contain at least 1 directory.");
        }
        
        

        for (int i =0; i < dirList.size();i++ ){
            String dir = dirList.get(i);
            String lastPath = lastPaths.get(i);
            String order = task.getOrder();
            ArrayList<String> files = listFiles(task,Paths.get(dir).normalize(),lastPath,order);
            // Sort the files if each directory
            
            
            if (order.equals("ALPHABETICAL"){
                Collections.sort(files);
            } else if(order.equals("ASCEND_MODIFIED") || order.equals("DESCEND_MODIFIED"){
                Collections.sort(files,new Comparator<String>(){
                    @Override
                    public int compare(String f1, String f2) {
                        try{
                            return getLastModifiedTime(f1).compareTo(getLastModifiedTime(f2));
                        } catch (IOException ex){
                            ex.printStackTrace();
                        }
                        return 0;
                    }
                });
                if (order.equals("DESCEND_MODIFIED"){ Collections.reverse(files); }
            }
            } else if (order.equals("ASCEND_CREATION") || order.equals("DESCEND_CREATION") ){
                Collections.sort(files,new Comparator<String>(){
                    @Override
                    public int compare(String f1, String f2) {
                        try{
                            return getCreationTime(f1).compareTo(getCreationTime(f2));
                        } catch (IOException ex){
                            ex.printStackTrace();
                        }
                        return 0;
                    }
                });
          
                if ( order.equals("DESCEND_CREATION") ) { Collections.reverse(files);}
            } else {
                throw new RuntimeException("Input a correct order");
            }
            
            // End of sort
            
            log.info("The files is " + files);
            
            allFiles.add(files);
            
        }
        
        int taskCount;
        // If the we upload only one directory, we set each file as a task.
        // In this case the max_threads must equal 1 to keep the file uploading order
        if (dirList.size() == 1){
            log.info("size==1");
            ArrayList<ArrayList<String>> oneFile = new ArrayList<ArrayList<String>> ();
            for(String f : allFiles.get(0)){
                ArrayList<String> file = new ArrayList<String> ();
                file.add(f);
                oneFile.add(file);
            }
            while (tagList.size()< oneFile.size()){
                tagList.add(tagList.get(0));
            }
            task.setFiles(oneFile);
            taskCount = oneFile.size();
        } else{
            task.setFiles(allFiles);
            taskCount = allFiles.size();
        }
        
        ArrayList<ColumnConfig> columns = new ArrayList<ColumnConfig>();
        //final String columnName = task.getColumnName();

        columns.add(new ColumnConfig("payload", STRING, config));
        columns.add(new ColumnConfig("tag", STRING, config));
        
        Schema schema = new SchemaConfig(columns).toSchema();


        //Schema schema = task.getColumns().toSchema();
          // number of run() method calls
        
        log.info("TASKCOUNT " + taskCount);

        return resume(task.dump(), schema, taskCount, control);
    }

    @Override
    public ConfigDiff resume(TaskSource taskSource,
            Schema schema, int taskCount,
            InputPlugin.Control control)
    {
        control.run(taskSource, schema, taskCount);
        return Exec.newConfigDiff();
    }

    @Override
    public void cleanup(TaskSource taskSource,
            Schema schema, int taskCount,
            List<TaskReport> successTaskReports)
    {
    }

    @Override
    public TaskReport run(TaskSource taskSource,
            Schema schema, int taskIndex,
            PageOutput output)
    {
        PluginTask task = taskSource.loadTask(PluginTask.class);
        
        ArrayList<String> files = task.getFiles().get(taskIndex);
        
        log.info("The files in the run:" + files);
        

        for (String file : files)
        {
            try
            {
                int nRead;
                byte[] data = new byte[chunkSize];
                FileInputStream dataIn  = new  FileInputStream(file);
                ByteArrayOutputStream buffer = new ByteArrayOutputStream();
                
                while ((nRead = dataIn.read(data, 0, data.length)) != -1) {
                    buffer.write(data, 0, nRead);
                    try (PageBuilder pageBuilder = new PageBuilder(Exec.getBufferAllocator(), schema, output))
                    {
                        pageBuilder.setString(0,buffer.toString());//Base64.encodeBase64String(buffer.toByteArray()));
                        pageBuilder.setString(1, tagList.get(taskIndex) + new File(file).getCanonicalPath() );
                        pageBuilder.addRecord();
                        buffer.flush();
                        pageBuilder.finish();
                    }
                }
            } catch (IOException ex){
                ex.printStackTrace();
            }
        }

        TaskReport taskReport = Exec.newTaskReport();
        taskReport.set("columns", schema.size());
        return taskReport;
    }

    @Override
    public ConfigDiff guess(ConfigSource config)
    {
        return Exec.newConfigDiff();
    }
    
    
    public ArrayList<String> listFiles(PluginTask task,Path pathPrefix,String lastPath,String order)
    {
        //Path pathPrefix = Paths.get(task.getPathPrefix()).normalize();
        final Path directory;
        final String fileNamePrefix;
        if (Files.isDirectory(pathPrefix)) {
            directory = pathPrefix;
            fileNamePrefix = "";
        } else {
            fileNamePrefix = pathPrefix.getFileName().toString();
            Path d = pathPrefix.getParent();
            directory = (d == null ? CURRENT_DIR : d);
        }

        //final ImmutableList.Builder<String> builder = ImmutableList.builder();
        final ArrayList<String> filesArray = new ArrayList<String>();
        try {
            log.info("Listing local files at directory '{}' filtering filename by prefix '{}'", directory.equals(CURRENT_DIR) ? "." : directory.toString(), fileNamePrefix);
            Files.walkFileTree(directory, new SimpleFileVisitor<Path>() {
                @Override
                public FileVisitResult preVisitDirectory(Path path, BasicFileAttributes attrs)
                {
                    if (path.equals(directory)) {
                        return FileVisitResult.CONTINUE;
                    } else if (lastPath != null && path.toString().compareTo(lastPath.substring(0, path.toString().length())) < 0) {
                        return FileVisitResult.SKIP_SUBTREE;
                    } else if (path.getFileName().toString().startsWith(".")) {
                        return FileVisitResult.SKIP_SUBTREE;
                    } else {
                        if (path.getFileName().toString().startsWith(fileNamePrefix)) {
                            return FileVisitResult.CONTINUE;
                        } else {
                            return FileVisitResult.SKIP_SUBTREE;
                        }
                    }
                }


                @Override
                public FileVisitResult visitFile(Path path, BasicFileAttributes attrs)
                {
                    if ( !lastPath.equals("") && order.equals("ALPHABETICAL") && path.toString().compareTo(lastPath) <= 0) {
                        return FileVisitResult.CONTINUE;
                    } else if (!lastPath.equals("") && order.equals("ASCEND_MODIFIED") && getLastModifiedTime(pah.toString()).compareTo(getLastModifiedTime(lastPath)) <= 0) {
                        return FileVisitResult.CONTINUE;
                    } else if (!lastPath.equals("") && order.equals("DESCEND_MODIFIED") && getLastModifiedTime(pah.toString()).compareTo(getLastModifiedTime(lastPath)) >= 0){
                        return FileVisitResult.CONTINUE;
                    } else if (!lastPath.equals("") && order.equals("ASCEND_CREATION") && getLastCreationTime(pah.toString()).compareTo(getLastCreationTime(lastPath)) <= 0){
                        return FileVisitResult.CONTINUE;
                    } else if (!lastPath.equals("") && order.equals("DESCEND_MODIFIED") && getLastCreationTime(pah.toString()).compareTo(getLastCreationTime(lastPath)) <= 0) {
                        return FileVisitResult.CONTINUE;
                    }
                    else if (path.getFileName().toString().startsWith(".")) {
                         return FileVisitResult.CONTINUE;
                    } else {
                        if (path.getFileName().toString().startsWith(fileNamePrefix)) {
                              filesArray.add(path.toString());
                        }
                        return FileVisitResult.CONTINUE;
                    }
               }
            });
        } catch (IOException ex) {
            throw new RuntimeException(String.format("Failed get a list of local files at '%s'", directory), ex);
        }
        //return builder.build();
        return filesArray;
    }

    
    
    // End
    
    public static FileTime getCreationTime(String filename) throws IOException{
      File file = new File(filename);
      Path p = Paths.get(file.getAbsolutePath());
      BasicFileAttributes view = Files.getFileAttributeView(p,BasicFileAttributeView.class).readAttributes();
      FileTime fileTime = view.creationTime();
      //System.out.println("The raw creation time of " +filename+  " is " + fileTime.toString());
      return fileTime;
    }
  
    public static FileTime getLastModifiedTime(String filename) throws IOException{
      File file = new File(filename);
      Path p = Paths.get(file.getAbsolutePath());
      BasicFileAttributes view = Files.getFileAttributeView(p,BasicFileAttributeView.class).readAttributes();
      FileTime fileTime = view.lastModifiedTime();
      //System.out.println("The raw last modified time of " +filename+  " is " + fileTime.toString());
      return fileTime;
    }
}
