package org.embulk.input.filename;

import com.google.common.collect.ImmutableList;
import java.util.stream.Stream;
import java.util.stream.Collectors;
import java.nio.file.attribute.BasicFileAttributes;
import java.nio.file.attribute.BasicFileAttributeView;
import java.nio.file.attribute.FileTime;
import java.util.Comparator;


import org.embulk.config.ConfigSource;
import org.embulk.config.ConfigDiff;
import org.embulk.test.EmbulkTests;
import org.embulk.spi.InputPlugin;
import org.embulk.spi.ParserPlugin;
import org.embulk.spi.OutputPlugin;
import org.embulk.spi.SchemaConfig;
import org.embulk.spi.ColumnConfig;
import org.junit.Rule;
import org.junit.Before;
import org.junit.Test;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.Files;
import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Arrays;
import java.util.ListIterator;


import static org.embulk.test.EmbulkTests.readSortedFile;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertThat;

public class TestFilenameInputPlugin
{

    public static FileTime getCreationTime(String filename) throws IOException{
        File file = new File(filename);
        Path p = Paths.get(file.getAbsolutePath());
        BasicFileAttributes view = Files.getFileAttributeView(p,BasicFileAttributeView.class).readAttributes();
        FileTime fileTime = view.creationTime();
        return fileTime;
    }

    public static FileTime getLastModifiedTime(String filename) throws IOException{
        File file = new File(filename);
        Path p = Paths.get(file.getAbsolutePath());
        BasicFileAttributes view = Files.getFileAttributeView(p,BasicFileAttributeView.class).readAttributes();
        FileTime fileTime = view.lastModifiedTime();
        return fileTime;
    }
  

    @Rule
    public TestHelper embulk = TestHelper.builder()
          .registerPlugin(InputPlugin.class,"filename",FilenameInputPlugin.class)
          .registerPlugin(OutputPlugin.class,"joinfile",JoinfileOutputPlugin.class)
          .build();
          
    
    @Test
    public void testOrderByModifiedTime() throws Exception{
    
        ConfigSource execConfig  = embulk.newConfig()
              .set("max_threads","1");
    
        Path path_src =  Paths.get("src/test/resources/testModifiedOrder");
        
        ArrayList<String> multi_dir = new ArrayList<String> ();
        multi_dir.add(path_src.toAbsolutePath().toString()+"/sample_");
        ConfigSource inConfig  = embulk.newConfig()
          .set("type","filename")
          .set("multi_dir",multi_dir)
          .set("order_by_modified_time","2");

        Path tmp = embulk.createTempDir();
        
        ConfigSource outConfig = embulk.newConfig()
          .set("type","joinfile")
          .set("sum_type","filename")
          .set("path_prefix",tmp.toString()+"/outputfile")
          .set("file_ext",".txt");
        
        TestHelper.RunResult res = embulk.runAllBuilder(execConfig,inConfig,outConfig);
        
        //Attention the readAllLines load all lines into memory, it is not recommanded to read a big file.
        List<String> lines = Files.readAllLines(Paths.get(tmp.toString()+"/outputfile.txt"));
        
        List<String> actual = Files.walk(path_src)
                                    .filter(Files::isRegularFile)
                                    .map(Path::toAbsolutePath)
                                    .map(Path::toString)
                                    .collect(Collectors.toList());
        
        
        Collections.sort(actual,new Comparator<String>(){
                    @Override
                    public int compare(String f1, String f2) {
                        try{
                            return getLastModifiedTime(f1).compareTo(getLastModifiedTime(f2));
                        } catch (IOException ex){
                            ex.printStackTrace();
                        }
                        return 0;
                    }
                });
          
        //System.out.println("The lines" + lines);
        //System.out.println("The actual" + actual);
        assertEquals(lines,actual);
    
        inConfig.set("order_by_modified_time","1");
        res = embulk.runAllBuilder(execConfig,inConfig,outConfig);
        lines = Files.readAllLines(Paths.get(tmp.toString()+"/outputfile.txt"));
        
        // We reverse the actual files
        Collections.reverse(actual);
        
        assertEquals(lines,actual);

    }
  
    
    
    @Test
    public void testTagList() throws Exception{

        ConfigSource execConfig  = embulk.newConfig()
              .set("max_threads","1");
    
        Path path_src =  Paths.get("src/test/resources/testDirList");
        
        // Be careful the name of the List should be multi_dir!
        List<String> multi_dir = Arrays.asList(path_src.toAbsolutePath().toString()+"/sample/sample_",path_src.toAbsolutePath().toString()+"/example/example_");
        List<String> multi_tag = Arrays.asList("hello","world");
        
        ConfigSource inConfig  = embulk.newConfig()
          .set("type","filename")
          .set("order_by_modified_time","2")
          .set("multi_dir",multi_dir)
          .set("multi_tag",multi_tag);
        System.out.println(inConfig);
          
        Path tmp = embulk.createTempDir();
        
        ConfigSource outConfig = embulk.newConfig()
          .set("type","joinfile")
          .set("sum_type","filename")
          .set("path_prefix",tmp.toString()+"/outputfile")
          .set("file_ext",".txt");
        
        TestHelper.RunResult res = embulk.runAllBuilder(execConfig,inConfig,outConfig);
        
        List<String> lines = Files.readAllLines(Paths.get(tmp.toString()+"/outputfile.txt"));
        
        //List<String> actual = Files.readAllLines(Paths.get(path_src+"/test.csv"));
        
        List<String> dir1 = Files.walk(Paths.get(path_src.toAbsolutePath().toString()+"/sample"))
                                    .filter(Files::isRegularFile)
                                    .map(Path::toAbsolutePath)
                                    .map(Path::toString)
                                    .collect(Collectors.toList());
                            
                            
        Collections.sort(dir1,new Comparator<String>(){
                    @Override
                    public int compare(String f1, String f2) {
                        try{
                            return getLastModifiedTime(f1).compareTo(getLastModifiedTime(f2));
                        } catch (IOException ex){
                            ex.printStackTrace();
                        }
                        return 0;
                    }
                });
        
        List<String> dir2 = Files.walk(Paths.get(path_src.toAbsolutePath().toString()+"/example"))
                                    .filter(Files::isRegularFile)
                                    .map(Path::toAbsolutePath)
                                    .map(Path::toString)
                                    .collect(Collectors.toList());
                            
                            
        Collections.sort(dir2,new Comparator<String>(){
                    @Override
                    public int compare(String f1, String f2) {
                        try{
                            return getLastModifiedTime(f1).compareTo(getLastModifiedTime(f2));
                        } catch (IOException ex){
                            ex.printStackTrace();
                        }
                        return 0;
                    }
                });
                
        
        for (ListIterator i = dir1.listIterator(); i.hasNext(); )
        {
            i.set(multi_tag.get(0) + i.next());
        }
        
        for (ListIterator i = dir2.listIterator(); i.hasNext(); )
        {
            i.set(multi_tag.get(1) + i.next());
        }

        
        
        dir1.addAll(dir2);
        //System.out.println(lines);
        //System.out.println(dir1);
        assertEquals(lines,dir1);

    }
  
    
    @Test
    public void testDirList() throws Exception{

        ConfigSource execConfig  = embulk.newConfig()
              .set("max_threads","1");
    
        Path path_src =  Paths.get("src/test/resources/testDirList");
        
        // Be careful the name of the List should be multi_dir!
        List<String> multi_dir = Arrays.asList(path_src.toAbsolutePath().toString()+"/sample/sample_",path_src.toAbsolutePath().toString()+"/example/example_");
        List<String> multi_tag = Arrays.asList("hello","world");
        
        ConfigSource inConfig  = embulk.newConfig()
          .set("type","filename")
          .set("order_by_modified_time","2")
          .set("multi_dir",multi_dir)
          .set("path_prefix","/home/chronos/user/Downloads/embulk-input-filename/src/test/resources/testDirList/example/example_");
        Path tmp = embulk.createTempDir();
        
        ConfigSource outConfig = embulk.newConfig()
          .set("type","joinfile")
          .set("sum_type","filename")
          .set("path_prefix",tmp.toString()+"/outputfile")
          .set("file_ext",".txt");
        
        TestHelper.RunResult res = embulk.runAllBuilder(execConfig,inConfig,outConfig);
        
        List<String> lines = Files.readAllLines(Paths.get(tmp.toString()+"/outputfile.txt"));
        
        List<String> dir1 = Files.walk(Paths.get(path_src.toAbsolutePath().toString()+"/sample"))
                                    .filter(Files::isRegularFile)
                                    .map(Path::toAbsolutePath)
                                    .map(Path::toString)
                                    .collect(Collectors.toList());
                            
                            
        Collections.sort(dir1,new Comparator<String>(){
                    @Override
                    public int compare(String f1, String f2) {
                        try{
                            return getLastModifiedTime(f1).compareTo(getLastModifiedTime(f2));
                        } catch (IOException ex){
                            ex.printStackTrace();
                        }
                        return 0;
                    }
                });
        
        List<String> dir2 = Files.walk(Paths.get(path_src.toAbsolutePath().toString()+"/example"))
                                    .filter(Files::isRegularFile)
                                    .map(Path::toAbsolutePath)
                                    .map(Path::toString)
                                    .collect(Collectors.toList());
                            
                            
        Collections.sort(dir2,new Comparator<String>(){
                    @Override
                    public int compare(String f1, String f2) {
                        try{
                            return getLastModifiedTime(f1).compareTo(getLastModifiedTime(f2));
                        } catch (IOException ex){
                            ex.printStackTrace();
                        }
                        return 0;
                    }
                });
        
        
        dir1.addAll(dir2);
        //System.out.println(lines);
        //System.out.println(dir1);
        assertEquals(lines,dir1);
    }
    
    
    @Test
    public void testContent() throws Exception{

        ConfigSource execConfig  = embulk.newConfig()
              .set("max_threads","1");
    
        Path path_src =  Paths.get("src/test/resources/data");
        
        ArrayList<String> multi_dir = new ArrayList<String> ();
        multi_dir.add(path_src.toAbsolutePath().toString()+"/test.csv");
        ConfigSource inConfig  = embulk.newConfig()
          .set("type","filename")
          .set("multi_dir",multi_dir)
          .set("parser",embulk.newConfig().set("type","none-bin"));
          
        Path tmp = embulk.createTempDir();
        
        ConfigSource outConfig = embulk.newConfig()
          .set("type","joinfile")
          .set("sum_type","content")
          .set("path_prefix",tmp.toString()+"/outputfile")
          .set("file_ext",".txt");
        
        TestHelper.RunResult res = embulk.runAllBuilder(execConfig,inConfig,outConfig);
        
        List<String> lines = Files.readAllLines(Paths.get(tmp.toString()+"/outputfile.txt"));
        
        List<String> actual = Files.readAllLines(Paths.get(path_src+"/test.csv"));
        //System.out.println("The lines " + lines);
        //System.out.println("The actual " + actual);
        assertEquals(actual,lines);
    }
    
    
}
